/* bcb.c 
**
** This file defines the functions used by the BCB program 
** runbcb.c.  These are adapted from the FORTRAN implementation 
** and are declared in bcb.h . The random number generation 
** libraries rngs.h and rvgs.h are also used. 
**
** The user should redefine the functions evalcon() and 
** evalobj() to evaluate the constraints and objective
** function for the user's problem.  By default, evalobj()
** evaluates the sphere function and evalcon() evaluates
** all constraints as 0.
**
** The user should also construct a parameter input file
** named bcbparams.dat which has the format

 POPSIZE     100
 NUMGENS     1000
 SIGMA_M     1.0
 SIGMA_R     4.0
 PENALTY1    1000.0
 PENALTY2    1000.0
 PARCHOOSE   5
 NUMDV       3
 UBOUNDS     200.0 20.0   2.0
 LBOUNDS     0.0    1.0   0.5

**
** Mike Weber
** 2/2000
*/

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include "bcb.h"
#include "rngs.h"
#include "rvgs.h"
#include "rvms.h"


/* cumulative distribution function for distributions */
/* not included in rvms library                       */
double cdfWeibull(double lambda, double kappa, double t)
{
  return 1.0 - exp(-pow((lambda*t), kappa));
}

double cdfLoglogistic(double lambda, double kappa, double t)
{
  return 1.0 - 1.0/(1.0 + pow(lambda*t,kappa)); 
}

double cdfExpPower(double lambda, double kappa, double t)
{
  return 1.0 - exp(1.0-exp(lambda*pow(t, kappa))); 
}

double cdfGompertz(double delta, double kappa, double t)
{
  return 1.0 - exp(-delta*(pow(kappa, t) - 1.0)/log(kappa));
}


/************************************************************************
** distance(p1, p2)
** returns the Euclidean distance between points p1 and p2
*************************************************************************/
double distance(struct SOLUTION p1, struct SOLUTION p2)
{
  double dist = 0.0;
  int i;
  for(i = 0; i < p1.numdv; i++){
    dist += pow( (p1.dv[i]-p2.dv[i]), 2.0);
  }
  return sqrt(dist);
}


/************************************************************************
** evalcon(pop)
** evaluate the constraints for each solution in the population
** THIS FUNCTION SHOULD BE REDEFINED BY THE USER TO FIT 
** THE USER'S PROBLEM.
*************************************************************************/
void evalcon(struct POPULATION *pop)
{
  int i;
  
  /* By default, set constraints to 0 */
  for(i=0; i<(*pop).popsize; i++){
    (*pop).points[i].maxcon = 0.0;
    (*pop).points[i].avgcon = 0.0;
  }
}



/************************************************************************
** evalobj(pop)
** evaluate the objective function for each solution in the population
** THIS FUNCTION SHOULD BE REDEFINED BY THE USER TO FIT 
** THE USER'S PROBLEM.
*************************************************************************/
void evalobj(struct POPULATION *pop, double data[NUMPOINTS], int model, int numpoints)
{
  int i, j;
  double time, tempplus, tempminus, maxks, dnplus, dnminus, ks;


  /* k-s statistic */
  for(j=0; j <(*pop).popsize; j++){ 
      maxks = 0.0;
      dnplus = 0.0;
      dnminus = 0.0;
      for(i=0; i < numpoints; i++){
	switch(model){
	case(0):  tempplus = (double)(i+1.0)/(double)numpoints - 
		    cdfExponential((*pop).points[j].dv[0], data[i]); 
	          tempminus = cdfExponential((*pop).points[j].dv[0], data[i]) 
				    - (double)(i)/(double)(numpoints); break; 
	case(1):  tempplus = (double)(i+1.0)/(double)numpoints - 
		    cdfNormal((*pop).points[j].dv[0], (*pop).points[j].dv[1], data[i]); 
	          tempminus = cdfNormal((*pop).points[j].dv[0], (*pop).points[j].dv[1], data[i]) 
				    - (double)(i)/(double)(numpoints); break; 
	case(2):  tempplus = (double)(i+1.0)/(double)numpoints - 
		    cdfWeibull((*pop).points[j].dv[0], (*pop).points[j].dv[1], data[i]); 
	          tempminus = cdfWeibull((*pop).points[j].dv[0], (*pop).points[j].dv[1], data[i]) 
				    - (double)(i)/(double)(numpoints); break;  
	case(3):  tempplus = (double)(i+1.0)/(double)numpoints - 
		    cdfLognormal((*pop).points[j].dv[0], (*pop).points[j].dv[1], data[i]); 
	          tempminus = cdfLognormal((*pop).points[j].dv[0], 
			      (*pop).points[j].dv[1], data[i]) - (double)(i)/(double)(numpoints); 
		  break; 
	case(4):  tempplus = (double)(i+1.0)/(double)numpoints - 
		    cdfLoglogistic((*pop).points[j].dv[0], (*pop).points[j].dv[1], data[i]); 
	          tempminus = cdfLoglogistic((*pop).points[j].dv[0], 
			      (*pop).points[j].dv[1], data[i]) - (double)(i)/(double)(numpoints); 
		  break; 
	case(5):  tempplus = (double)(i+1.0)/(double)numpoints - 
		    cdfGompertz((*pop).points[j].dv[0], (*pop).points[j].dv[1], data[i]); 
	          tempminus = cdfGompertz((*pop).points[j].dv[0], 
			      (*pop).points[j].dv[1], data[i]) - (double)(i)/(double)(numpoints); 
		  break;
	case(6):  tempplus = (double)(i+1.0)/(double)numpoints - 
		    InGamma((*pop).points[j].dv[1], 1.0/(*pop).points[j].dv[0]*data[i]); 
	          tempminus = InGamma((*pop).points[j].dv[1], 1.0/(*pop).points[j].dv[0]*data[i]) 
		              - (double)(i)/(double)(numpoints); 
		  break;
	case(7):  tempplus = (double)(i+1.0)/(double)numpoints - 
		    cdfExpPower((*pop).points[j].dv[0], (*pop).points[j].dv[1], data[i]); 
	          tempminus = cdfExpPower((*pop).points[j].dv[0], (*pop).points[j].dv[1], data[i]) 
		              - (double)(i)/(double)(numpoints); 
		  break;

	}        


	if(tempplus > dnplus) dnplus = tempplus;
	if(tempminus > dnminus) dnminus = tempminus;
  
      }
        
      if(dnplus > dnminus)
       (*pop).points[j].obj = dnplus;
       else
       (*pop).points[j].obj = dnminus;
      if((*pop).points[j].obj == 0)
	(*pop).points[j].obj = 0.999;
     
  }  
}



/*************************************************************************
** fitness(pop, param)
** calculates the fitness of each individual in the population by
** penalizing the objective by some factor times the maximum constraint, 
** if the maximum constraint is greater than 0.
**************************************************************************/
void fitness(struct POPULATION *pop, struct SOLUTION *best, struct PARAMLIST *par)
{
  int i;
  double pen;
  
  if((*best).maxcon <= 0.0){
    pen = (*par).pen1;
  } else {
    pen = (*par).pen2;
  }
  
  for(i=0; i < (*pop).popsize; i++){
    if((*pop).points[i].maxcon <= 0.0){
       (*pop).points[i].fitness = (*pop).points[i].obj;
    } else {
      (*pop).points[i].fitness = (*pop).points[i].obj +
	                         pen * (*pop).points[i].maxcon;
    }
  }
}


/*************************************************************************
** fitnesssrt(pop)
** sorts the solutions in the population by
** 1. objective function value
** 2. max constraint 
**************************************************************************/
void fitnesssrt(struct POPULATION *pop)
{
  int used1[MAXPOPSIZE], used2[MAXPOPSIZE];
  int i, j, index1, index2;
  double bigobj, bigmaxcon;

  bzero(used1, (*pop).popsize*sizeof(int));
  bzero(used2, (*pop).popsize*sizeof(int));
  
  for(i=0; i<(*pop).popsize; i++){
    bigobj =  BIGM;  bigmaxcon =  BIGM;
    for(j = 0; j<(*pop).popsize; j++){
      if((*pop).points[j].obj < bigobj && used1[j] == 0){
	bigobj = (*pop).points[j].obj;
	index1 = j;
      }
      if((*pop).points[j].maxcon < bigmaxcon && used2[j] == 0){
	bigmaxcon = (*pop).points[j].maxcon;
	index2 = j;
      }
    }
    used1[index1] = 1;
    used2[index2] = 1;
    (*pop).sortbyobj[i] = index1;
    (*pop).sortbymax[i] = index2;
  }
}


/************************************************************************
** initpop(pop, param)
** intializes the decision variables for the solutions in a population 
*************************************************************************/
void initpop(struct POPULATION *pop, struct PARAMLIST *par)
{
  int i, j;
  (*pop).popsize = (*par).popsize;
  for(i=0; i<(*pop).popsize; i++){
    (*pop).points[i].numdv = (*par).numdv; 
   

    for(j=0; j<(*par).numdv; j++){
        (*pop).points[i].dv[j] = Random() 
                               * ( (*par).ub[j] - (*par).lb[j] )
                               + (*par).lb[j];
    }
  }
}


/*************************************************************************
** poolpops(pooled, pop1, pop2)
** combines pop1 and pop2 into a pooled population 
**************************************************************************/
void poolpops(struct POPULATION *pool, struct POPULATION *pop1, 
              struct POPULATION *pop2)
{
  (*pool).popsize = 2 * (*pop1).popsize;
  memcpy((*pool).points, (*pop1).points, 
         (*pop1).popsize*sizeof(struct SOLUTION));
  memcpy(&((*pool).points[(*pop1).popsize]), (*pop2).points, 
	 (*pop1).popsize*sizeof(struct SOLUTION));
}  


/*************************************************************************
** read_param(paramlist, file)
** reads the input file of bcb paramters, in the format given above
**************************************************************************/
void read_param(struct PARAMLIST *par)
{
  FILE *f;
  char junk[20];
  int i;

  f = fopen("bcbparams.dat", "r");
  fscanf(f, "%s %ld", junk, &((*par).popsize));
  fscanf(f, "%s %ld", junk, &((*par).numgens));
  fscanf(f, "%s %lf", junk, &((*par).sigma_m));
  fscanf(f, "%s %lf", junk, &((*par).sigma_r));
  fscanf(f, "%s %lf", junk, &((*par).pen1));
  fscanf(f, "%s %lf", junk, &((*par).pen2));
  fscanf(f, "%s %d", junk, &((*par).parchoose));
  fscanf(f, "%s %ld", junk, &((*par).numdv));
  fscanf(f, "%s", junk);
  for(i=0; i<(*par).numdv; i++){
    fscanf(f, "%lf", &((*par).ub[i]));
  }
  fscanf(f, "%s", junk); 
  for(i=0; i<(*par).numdv; i++){
    fscanf(f, "%lf", &((*par).lb[i]));
  }
  fclose(f);
}



/*************************************************************************
** recombine(newpop, oldpop)
** mates pairs of solutions in oldpop to produce a new 
** population of children
**
** For parent selection, Baker's 1-criterion and 2-criteria are
** implemented as PARCHOOSE=5 and PARCHOOSE=6
**************************************************************************/
void recombine(struct POPULATION *newpop, struct POPULATION *oldpop, 
               struct PARAMLIST *par)
{
  struct SOLUTION parent1, parent2, child;
  struct SOLUTION mvparent1, mvparent2, posparent;
  double parsum1, parsum2, negfit, avgfit1, avgfit2;
  double sumexpoff1, sumexpoff2, scale1, scale2, mean;
  double expoff1[MAXPOPSIZE], expoff2[MAXPOPSIZE];
  double expcopy[MAXPOPSIZE], gen[MAXDV];
  double s[MAXDV], ctheta[MAXDV], stheta[MAXDV], r;
  double dist, radius, sum;
  long p1, p2, picked;
  long shuf1[MAXPOPSIZE], shuf2[MAXPOPSIZE];
  long plist1[2*MAXPOPSIZE], plist2[2*MAXPOPSIZE];
  double num, alpha, largest, lpp;
  int i, j, k, temp, parfl, flag;
  double temp1, temp2;

  /* set up bins for selection using Baker's */
  if((*par).parchoose == 5 || (*par).parchoose == 6 ){
    
    negfit = 1.0;
    if((*oldpop).points[(*oldpop).sortbymax[0]].maxcon <= 0.0){
      negfit = (*oldpop).points[(*oldpop).sortbymax[0]].maxcon;
    }
    
    parsum1 = parsum2 = 0.0;
    for(i=0; i<(*oldpop).popsize; i++){
      parsum1 += (*oldpop).points[(*oldpop).sortbyobj[i]].obj;
      parsum2 += (*oldpop).points[(*oldpop).sortbymax[i]].maxcon;
      if(negfit < 0.0){
	parsum2 += (100.0-negfit); 
      }
    }
    assert( (*oldpop).popsize != 0);
    avgfit1 = parsum1/(double)(*oldpop).popsize;
    avgfit2 = parsum2/(double)(*oldpop).popsize;
    
    sumexpoff1 = sumexpoff2 = 0.0;
    for(i=0; i<(*oldpop).popsize; i++){
      assert((*oldpop).points[(*oldpop).sortbyobj[i]].obj != 0);
      expoff1[i] = avgfit1/(*oldpop).points[(*oldpop).sortbyobj[i]].obj;
      if(negfit <= 0.0){
	assert(( (*oldpop).points[(*oldpop).sortbymax[i]].maxcon
                              + 100.0 - negfit) != 0);
	expoff2[i] = avgfit2/( (*oldpop).points[(*oldpop).sortbymax[i]].maxcon
                              + 100.0 - negfit); 
      } else {
	assert((*oldpop).points[(*oldpop).sortbymax[i]].maxcon != 0);
	expoff2[i] = avgfit2/(*oldpop).points[(*oldpop).sortbymax[i]].maxcon;
      }
      sumexpoff1 += expoff1[i];
      sumexpoff2 += expoff2[i];
    }
    
    assert(sumexpoff1 != 0); 
    scale1 = (double)((*oldpop).popsize)/sumexpoff1;
    scale2 = (double)((*oldpop).popsize)/sumexpoff2;
    for(i=0; i<(*oldpop).popsize; i++){
      expoff1[i] *= scale1;
      expoff2[i] *= scale2;
    }
  }


  /* Do the parent selection using 1-criterion Baker's */
  if((*par).parchoose == 5){
    for(i=0; i<(*oldpop).popsize; i++){ 
      expcopy[i] = expoff1[i];
      shuf1[i] = i;
    }
    
    for(i=0; i<(*oldpop).popsize; i++){  
      j = Equilikely(i, (*oldpop).popsize-1);
      temp = shuf1[j];
      shuf1[j] = shuf1[i];
      shuf1[i] = temp;
    }

    for(i=0; i<(*oldpop).popsize; i++){
      expoff1[i] = expcopy[shuf1[i]]*2.0;
    }
    
    num = Random();
    picked = 1;
    for(i=0; i<(*oldpop).popsize; i++){
      num += expoff1[i];
      while(picked < num  &&  picked <= 2*(*oldpop).popsize){
	plist1[picked-1] = shuf1[i];
	picked++;
      }
    }
  }
  
  if((*par).parchoose == 6){
    for(i=0; i<(*oldpop).popsize; i++){ 
      shuf1[i] = i;
      shuf2[i] = 1;
    }
    
    for(i=0; i<(*oldpop).popsize; i++){  
      j = Equilikely(i, (*oldpop).popsize-1);
      temp = shuf1[j];
      shuf1[j] = shuf1[i];
      shuf1[i] = temp;
      j = Equilikely(i, (*oldpop).popsize-1);
      temp = shuf2[j];
      shuf2[j] = shuf2[i];
      shuf2[i] = temp;
    }


    for(i=0; i<(*oldpop).popsize; i++){
      expcopy[i] = expoff1[i];
    }
    for(i=0; i<(*oldpop).popsize; i++){
      expoff1[i] = expcopy[shuf1[i]];
    }
    for(i=0; i<(*oldpop).popsize; i++){
      expcopy[i] = expoff2[i];
    }
    for(i=0; i<(*oldpop).popsize; i++){
      expoff2[i] = expcopy[shuf2[i]];
    } 

    
    num = Random();
    picked = 1;
    for(i=0; i<(*oldpop).popsize; i++){
      num += expoff1[i];
      while(picked < num  &&  picked <= (*oldpop).popsize){
	plist1[picked-1] = shuf1[i];
	picked++;
      }
    }
    num = Random();
    picked = 1;
    for(i=0; i<(*oldpop).popsize; i++){
      num += expoff2[i];
      while(picked < num  &&  picked <= (*oldpop).popsize){
	plist2[picked-1] = shuf2[i];
	picked++;
      }
    } 
  }


  /* Now choose pairs of parents and do recombination */
  if((*par).parchoose ==5 || (*par).parchoose ==6){
   (*newpop).popsize = (*oldpop).popsize;
   for(i=0; i<(*oldpop).popsize; i++){

    if((*par).parchoose == 5){
      p1 = (*oldpop).sortbyobj[plist1[i]];
      p2 = (*oldpop).sortbyobj[plist1[i+(*oldpop).popsize]];
    } else if((*par).parchoose == 6){
      p1 = (*oldpop).sortbyobj[plist1[i]];
      p2 = (*oldpop).sortbyobj[plist2[i]];
    }
    memcpy(&parent1, &((*oldpop).points[p1]), sizeof(struct SOLUTION));
    memcpy(&parent2, &((*oldpop).points[p2]), sizeof(struct SOLUTION));
    
   
    assert(parent1.fitness+parent2.fitness != 0);
    mean = parent1.fitness/(parent1.fitness+parent2.fitness);
    alpha = Normal(mean, (*par).sigma_m); 
    child.numdv = parent1.numdv;
    do{
     flag = 0;
     for(j=0; j<child.numdv; j++){
      child.dv[j] = (1.0-alpha)*parent1.dv[j] + alpha*parent2.dv[j];
      if(child.dv[j] > (*par).ub[j] || child.dv[j] < (*par).lb[j]){
	flag=1;
      }
     }
     alpha = 0.5 - (0.5-alpha)/1.2;
    }while(flag);


    
    
    memcpy(&mvparent1, &parent1, sizeof(struct SOLUTION));
    memcpy(&mvparent2, &parent2, sizeof(struct SOLUTION));
    for(j=0; j<parent1.numdv; j++){
      mvparent1.dv[j] -= child.dv[j];
      mvparent2.dv[j] -= child.dv[j];
    }
    memcpy(&posparent, &mvparent1, sizeof(struct SOLUTION));
    parfl = 1;


  
    /* scale each design variable to the largest */
    largest = -99999.99;
    for(j=0; j<(*par).numdv; j++){
       if(posparent.dv[j] > largest){
	 largest = posparent.dv[j];
       }
    }
    lpp = 0.0;
    for(j=0; j<posparent.numdv; j++){
      if(posparent.dv[j] == 0.0 || largest==0){
	s[j] = 1.0;
      } else {
	s[j] = largest/posparent.dv[j];
      }
      posparent.dv[j] = largest;
      lpp = lpp + posparent.dv[j]*posparent.dv[j];
    }
    lpp = sqrt(lpp);


    for(j=2; j<=(*par).numdv; j++){
      k = (*par).numdv - j + 1;
      r = pow(posparent.dv[0], 2.0) + pow(posparent.dv[k], 2.0);
      r = sqrt(r);
      if(r == 0.0){
	ctheta[k] = 1.0;
	stheta[k] = 0.0;
      } else {
	ctheta[k] = posparent.dv[0]/r;
	stheta[k] = posparent.dv[k]/r;
      }
      posparent.dv[0] = ctheta[k]*posparent.dv[0] + stheta[k]*posparent.dv[k];
      posparent.dv[k] = 0.0;
    }
    

    dist = distance(parent1, parent2);
    radius = fabs(Normal(0.0, dist*(*par).sigma_r)); 
    gen[0] = 0.0;
    sum = 0.0;
    for(j=1; j<(*par).numdv; j++){
      gen[j] = Normal(0.0, 1.0);
      sum += gen[j]*gen[j];
    }
    sum = sqrt(sum);
    assert(sum != 0);
    for(j=1; j<(*par).numdv; j++){
      gen[j] *= radius/sum;
    }


    for(j=1; j<(*par).numdv; j++){
      temp1 = ctheta[j]*gen[0] - stheta[j]*gen[j];
      temp2 = stheta[j]*gen[0] + ctheta[j]*gen[j];
      gen[0] = temp1;
      gen[j] = temp2;
    }
    for(j=0; j<(*par).numdv; j++){
      assert(s[j] != 0);
      gen[j] = gen[j]/s[j];
    }

    for(j=0; j<(*par).numdv; j++){
      (*newpop).points[i].dv[j] = gen[j] + child.dv[j];
    }


    for(j=0; j<(*par).numdv; j++){
      if((*newpop).points[i].dv[j] > (*par).ub[j]){
	(*newpop).points[i].dv[j] = (*par).ub[j];
      }
      if((*newpop).points[i].dv[j] < (*par).lb[j]){
	(*newpop).points[i].dv[j] = (*par).lb[j];
      }
    }
    (*newpop).points[i].numdv = (*par).numdv;    
   }
  }
}



/*************************************************************************
** select(new, pooled)
** selects the most fit solutions in the pooled population to survive
**************************************************************************/
void g_select(struct POPULATION *new, struct POPULATION *pooled)
{
  int i, j, index;
  long localfit[2*MAXPOPSIZE];
  double bigfit;

  for(j=0; j < (*pooled).popsize; j++){
    localfit[j] = j;
  }

  for(i=0; i< (*new).popsize; i++){
    bigfit =  BIGM;
    for(j=i; j < (*pooled).popsize; j++){
      if((*pooled).points[localfit[j]].fitness < bigfit ){
	bigfit = (*pooled).points[localfit[j]].fitness;
	index = j;
      }
    }
    memcpy( &((*new).points[i]),&((*pooled).points[localfit[index]]),
            sizeof(struct SOLUTION)); 
    localfit[index] = localfit[i];
  }
}


/*************************************************************************
** zbest()
** determines the best solution found so far, based on fitness values
** and updates the information if a new best solution is found
** prints objective and constraint values for best solution
**************************************************************************/
void zbest(struct POPULATION *pop, struct SOLUTION *best)
{
  if((*pop).points[0].fitness < (*best).fitness){
    memcpy(best, &(*pop).points[0], sizeof(struct SOLUTION));
  }
  /*  printf("Best Solution\n"); */
/*    printf("-------------\n"); */
/*    printf("Objective Value: %e\n", (*best).obj); */
/*    printf("Max. Constraint: %lf\n", (*best).maxcon); */
/*    printf("Avg. Constraint: %lf\n", (*best).avgcon); */
/*    printf("Best lambda:     %lf\n\n\n", (*best).dv[0]); */
}
