/*  runbcb.c 
**
**  This program makes use of the functions and structures
**  declared in bcb.h, defined in bcb.c, to implement the
**  BCB algorithm.  The random number generation library
**  rngs.h is also used.
**
**  Mike Weber, 3/2000
*/

#include "bcb.h"
#include "rngs.h"
#include <math.h>
#include <stdlib.h>
#include <stdio.h>

void read_data(double data[NUMPOINTS], int *numpoints)
{
  int i;
  *numpoints = 0;
  while(!feof(stdin)){
    scanf("%lf", &(data[*numpoints]));
    (*numpoints)++;
  }
  (*numpoints)--;
}

void print_header()
{
  system("clear\n");
  printf("\n\nThis program is fitting the input data to seven common\n");
  printf("distribution models for failure times.  An optimization\n");
  printf("heuristic is being used to identify near-optimal parameters\n");
  printf("that minimize the K-S statistic for each model.\n");
  printf("\n\nModel\t\tK-S\t\tScale Param\t\tShape Param\n");
  printf(    "-----\t\t---\t\t-----------\t\t-----------\n");
}

void print_results(int model, struct SOLUTION bestsol)
{
  switch(model){
      case(0):
	printf("Exponential\t%.8lf\t%.8lf\t\t----\n", bestsol.obj,bestsol.dv[0]); break;
      case(1):
	printf("Normal\t\t%.8lf\t%.8lf\t\t%.8lf\n", bestsol.obj, 
	       bestsol.dv[0], bestsol.dv[1]); break;
      case(2):
	printf("Weibull\t\t%.8lf\t%.8lf\t\t%.8lf\n", bestsol.obj, 
	       bestsol.dv[0], bestsol.dv[1]); break;
      case(3):
	printf("Lognormal\t%.8lf\t%.8lf\t\t%.8lf\n", bestsol.obj, 
	       bestsol.dv[0], bestsol.dv[1]); break;
      case(4):
	printf("Log logistic\t%.8lf\t%.8lf\t\t%.8lf\n", bestsol.obj, 
	       bestsol.dv[0], bestsol.dv[1]); break;

      case(5):
	printf("Gompertz\t%.8lf\t%.8lf\t\t%.8lf\n", bestsol.obj, 
	       bestsol.dv[0], bestsol.dv[1]); break;
      case(6):
	printf("Gamma\t\t%.8lf\t%.8lf\t\t%.8lf\n", bestsol.obj, 
	       bestsol.dv[0], bestsol.dv[1]); break;
      case(7): 
	printf("Exp. Power\t%.8lf\t%.8lf\t\t%.8lf\n", bestsol.obj, 
	       bestsol.dv[0], bestsol.dv[1]); break;
      }
}

int main()
{
  struct POPULATION pop1, pop2, ppop;  /* current, children, and pooled populations */
  struct SOLUTION bestsol;             /* the best solution found so far            */
  struct PARAMLIST parlist;            /* the algorithm and problem parameters      */
  double data[NUMPOINTS];
  int i, model, numpoints;
  int minmodel;
  double minks;
  
  PutSeed(387967389);                  /* initialize the random number generator    */
  read_param(&parlist);                /* read in algortihm and problem parameters  */
  read_data(data, &numpoints);         /* read in list of input data                */
  print_header();                     /* print the header for program output       */
      
  minks = 1.0;
  for(model=0; model < 8; model++){
      bestsol.obj = BIGM;                  /* initialize the best solution variable     */
      bestsol.fitness = BIGM;              
      bestsol.maxcon = 0.0;

      initpop(&pop1, &parlist);               /* randomly create and evluate an initial */
      evalobj(&pop1, data, model, numpoints); /* population uniformly over the feasible */
      evalcon(&pop1);                         /* region                                 */
      fitness(&pop1, &bestsol, &parlist);     /*                                        */

      for(i=0; i < parlist.numgens; i++){     /* perform several generations of BCB     */
	fitnesssrt(&pop1);                    /* sort the current population by fitness */
	recombine(&pop2, &pop1, &parlist);    /* do recombination and store children in */ 
					      /* a new population                       */
	evalobj(&pop2, data,model,numpoints); /* evaluate the children                  */
	evalcon(&pop2); 
	fitness(&pop2, &bestsol, &parlist);
	poolpops(&ppop, &pop1, &pop2);        /* pool the parents and children together */
	g_select(&pop1, &ppop);               /* select survivors from the pooled pop   */
	zbest(&pop1, &bestsol);               /* update and report best solution so far */ 
      }
      print_results(model, bestsol);         /* print the results for one model */ 
      if(bestsol.obj < minks){
	minks = bestsol.obj;
	minmodel = model;
      }
  }
  return 0;
}
